#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-

"""

Architecture
------------

::

    +-----------------------------------------+
    |                                         |
    |  gulik process                          |       collector processes
    |                                         |
    |  (main thread)    (monitor threads)     |
    |   Gulik----------- CPUMonitor     <-----ø-----> CPUCollector
    |     |         | `- MemoryMonitor  <-----ø-----> MemoryCollector
    |     |         |`-- NetworkMonitor <-----ø-----> NetworkCollector
    |     |         `--- …              <-----ø-----> …
    |     |                   |               |
    |     |                   |               |
    |     `-- visualizers     |  (visualizers |
    |           | `- Arc      |   access      |
    |           |`-- Plot   <´    monitors)   |
    |           `--- …                        |
    |                                         |
    +-----------------------------------------+

In ``gulik``, there is one central :class:`Gulik` object.
It manages :ref:`monitor`-:ref:`collector` pairs and :ref:`visualizer`\s.

Visualizers use the :func:`monitors.Monitor.normalize` and :func:`monitors.Monitor.caption`
functions to utilize the collected data.

Communication between the :ref:`monitor`\s within the gulik process and
:ref:`collector` processes is done via queues. Every monitor/collector
pair shares two queues. One "update queue" that monitors use to send update
requests to collectors and one "data queue" that collectors use to send the
next datapoint to their respective monitor.

Box model
---------

::

    +-----------------------------------------------------------+   ↑
    |                                                           |   |
    |                       margin                              |   |
    |                                                           |   |
    |   +---------------------------------------------------+   |   |
    |   |                                                   |   |   
    |   |                   padding                         |   |   h
    |   |                                                   |   |   e
    |   |   +-----------------------------------+  ↑        |   |   i
    |   |   |                                   |  | inner  |   |   g
    |   |   |       inner drawing region        |  | height |   |   h
    |   |   |                                   |  |        |   |   t
    |   |   +-----------------------------------+  ↓        |   |   
    |   |                                                   |   |   |
    |   |   ←------------ inner width ----------→           |   |   |
    |   |                                                   |   |   |
    |   +---------------------------------------------------+   |   |
    |                                                           |   |
    +-----------------------------------------------------------+   ↓

    ←--------------------------- width -------------------------→

``gulik``\s box model is pretty similar to the one CSS has, with
one important distinction: margins and paddings are included in
``width`` and ``height`` in order to make layouting easier.

Concepts
--------

.. _visualizer:

visualizer
^^^^^^^^^^

Visualizers are instances of any subclass of :class:`visualizers.Visualizer`.

A visualizer is assigned a :ref:`monitor` and a list of :ref:`element`\s.

:class:`Gulik` will periodically call all visualizers ``update`` methods
(see source of :func:`visualizers.Visualizer.update`).

What exactly happens in the ``update`` function of a visualizer differs
between the different classes, but usually it queries the instance-assigned
monitor for the most recent data about its :ref:`element`\s by calling
:func:`monitor.Monitor.normalize` and then does some drawing on the passed
:class:`cairo.Context` to visualize the data in some manner.

Currently, there are 7 built-in visualizers:
    * :class:`visualizers.Text`
    * :class:`visualizers.Rect`
    * :class:`visualizers.Arc`
    * :class:`visualizers.Plot`
    * :class:`visualizers.MirrorRect`
    * :class:`visualizers.MirrorArc`
    * :class:`visualizers.MirrorPlot`

You are, however, welcome to implement your own visualizers by subclassing
:class:`visualizers.Visualizer` and overriding :func:`visualizers.Visualizer.draw`.

.. _monitor:

monitor
^^^^^^^

Monitors are instances of any subclass of :class:`monitors.Monitor`,
which itself is a subclass of :class:`multithreading.Thread`.

Every monitor acts as one half of a monitor-:ref:`collector` pair,
each of which collects and transforms data on a specific :ref:`component`.

The monitors responsibility in this pair is to take data from the collector
and offer it in a form that is usable by :ref:`visualizer`\s. This is mainly
done through the functions :func:`monitors.Monitor.normalize` and :func:`monitors.Monitor.caption`.

Currently, there are 6 built-in monitors:
    * :class:`monitors.CPUMonitor`
    * :class:`monitors.MemoryMonitor`
    * :class:`monitors.NetworkMonitor`
    * :class:`monitors.BatteryMonitor`
    * :class:`monitors.DiskMonitor`
    * :class:`monitors.NetdataMonitor`

.. _collector:

collector
^^^^^^^^^

Collectors are instances of any subclass of :class:`collectors.Collector`,
which itself is a subclass of :class:`multiprocessing.Process`.

Every collector acts as one half of a :ref:`monitor`-collector pair.

The collectors responsibility in this pair is to collect system usage
data and send them to its associated :ref:`monitor`.

.. _component:

component
^^^^^^^^^

A string identifying a data source.

Valid values are:
    * ``'cpu'``
    * ``'memory'``
    * ``'network'``
    * ``'battery'``
    * ``'disk'``

Besides that, ``'netdata-<hostname>'`` is valid, but only if ``<hostname>``
exists in the ``NETDATA_HOSTS`` configuration option.

.. _element:

element
^^^^^^^

A string identifying a (sub) element of a data source.
Valid values are defined within the respective :ref:`monitor`\s.

.. _alignment:

alignment
^^^^^^^^^

A string in the shape of ``<x-align>_<y-align>``, where:
    * ``<x-align>`` can be any of ``left``, ``center``, ``right``
    * ``<y-align>`` can be any of ``top``, ``center``, ``bottom``

alignments are used both for text positioning relative to its respective
allowed borders as well as positioning captions within :ref:`visualizer`\s.

.. _caption-description:

caption description
^^^^^^^^^^^^^^^^^^^

A dictionary describing a caption to be rendered by a :ref:`visualizer`.

Required items:
    * ``text``: The text to be rendered


Optional items:
    * ``position``: Either an ``(x,y)`` coordinate tuple or an :ref:`alignment`
    * ``align``: An :ref:`alignment`
    * ``font_size``: A number (int or float) denoting vertical font size in pixels
    * ``font_weight``: A string. ``Normal``, ``Bold``, ``Light``, etc.
    * ``operator``: One of cairos blending operators (See :class:`cairo.Operator`)


.. _pattern:

pattern
^^^^^^^

A function taking one :class:`helpers.Color` as parameter returning a cairo surface for
use as fill. See :func:`patterns.stripe45` for an example.


.. _palette:

palette
^^^^^^^

A function taking one :class:`helpers.Color` and one `int` parameter returning a
``list`` of :class:`helpers.Color` objects with its length being equal to the passed
`int` parameter.

.. note:: :func:`palettes.hue` and :func:`palettes.value` have extra parameters
    you won't be able to use without wrapping them in :func:`functools.partial`
    first!

.. _combination:

combination
^^^^^^^^^^^

A string denoting how multiple :ref:`element`\s are displayed within a :ref:`visualizer`.

Valid values are:
    * ``separate``: separate elements visually 
    * ``cumulative``: show values cumulatively, assume data is normalized for that (i.e. all values added max out at 1.0)
    * ``cumulative_force``: like ``cumulative``, but assumes every single value can go up to 1.0


"""

#__all__ = ['Gulik', 'Visualizer']

import os

# get version from _version file
from gulik._version import version as __version__

import sys
import math
import time
import random
import signal
import collections
import functools
import threading
import multiprocessing
import setproctitle
import psutil
import cairo
import gi

gi.require_version('Gtk', '3.0')
gi.require_version('PangoCairo', '1.0') # not sure if want
from gi.repository import Gtk, Gdk, GLib, Pango, PangoCairo

from . import helpers
from . import palettes
from . import patterns
from . import collectors
from . import monitors
from . import visualizers
from . import netdata

# add a few aliases to make configs a little less verbose
Operator = cairo.Operator
Color = helpers.Color
Box = helpers.Box

PAGESIZE = os.sysconf('SC_PAGESIZE')


##class PeriodicCall(threading.Thread):
#
#    """ Periodically forces a window to redraw """
#
#    def __init__(self, target, hz):
#        
#        super(PeriodicCall, self).__init__()
#        self.daemon = True
#
#        self.target = target
#        self.interval = 1.0/hz
#
#
#    def run(self):
#
#        while True: # This thread will automatically die with its parent because of the daemon flag
#
#            self.target()
#            time.sleep(self.interval)

try:
    _default_height = Gdk.Screen().get_default().get_height()
except Exception: # to avoid setup failures in sandboxes without graphical environment
    _default_height = 0

DEFAULTS = {
    'FPS': 1,
    'WIDTH': 200,
    'HEIGHT': _default_height,
    'X': 0,
    'Y': 0,
    'NETDATA_HOSTS': [],
    'NETDATA_RETRY': 5,
    
    'BSD_ACCURATE_MEMORY': False, # use inaccurate but cheaper memory collection on bsd

    # styling stuff below this

    'MARGIN': 5,
    'PADDING': 5,
    'PADDING_BOTTOM': 40, # space for 2-rows of autolegend

    'FONT': 'Orbitron',
    'FONT_WEIGHT': 'Light',
    'FONT_SIZE': 10,

    'COLOR_WINDOW_BACKGROUND': helpers.Color(0.05, 0.05, 0.05, 0.8),
    'COLOR_BACKGROUND': helpers.Color(1,1,1, 0.1), # background for visualizers
    'COLOR_FOREGROUND': helpers.Color(0.5, 1, 0, 0.6),
    'COLOR_CAPTION': helpers.Color(1,1,1, 0.6),
    #'COLOR_CAPTION_MINOR': helpers.Color(1,1,1, 0.3),
        
    'PALETTE': functools.partial(palettes.hue, distance=-120), # mhh, curry…
    'PATTERN': patterns.stripe45,

    'CAPTION_PLACEMENT': 'inner', # allow captions to be properly centered in the inner region of visualizers, as opposed to 'padding'
    #'CAPTION_PLACEMENT': 'padding', # allow captions to be placed within paddings, as opposed to 'inner'

    'LEGEND': True,
    'LEGEND_ORDER': 'normal', # other valid value: 'reverse'
    'LEGEND_SIZE': 20, # not font size, but height of one legend cell, including margin and padding.
    'LEGEND_PLACEMENT': 'padding',
    'LEGEND_MARGIN': 2.5,
    #'LEGEND_PADDING': 2.5,
    'LEGEND_PADDING': 0,

    'OPERATOR': Operator.OVER,

    # class-specific default styles
    
    'PADDING_TEXT': 0, # otherwise text will be tiny

    'PADDING_RECT': 5,
    'FONT_SIZE_RECT': 14,
    'FONT_WEIGHT_RECT': 'Bold',
    #'COLOR_RECT_CAPTION': helpers.Color(1,1,1, 1),
    'PATTERN_RECT': None,
    'OPERATOR_RECT_CAPTION': Operator.DIFFERENCE, # grants visibility no matter how much of the rect is filled
    'CAPTION_PLACEMENT_RECT': 'inner', # to get padding around cut-out

    'PATTERN_ARC': None,
    'PATTERN_MIRRORARC': None,

    'AUTOSCALE': False,
    'LINE': True,
    'MARKERS': False,
    'GRID': True,
}

## Stuff I'd much rather do without a huge dependency like gtk ##
class Window(Gtk.Window):

    def __init__(self):

        super(Window, self).__init__()

        self.set_title('gulik')
        self.set_role('gulik')
        self.stick() # show this window on every virtual desktop

        self.set_app_paintable(True)
        self.set_type_hint(Gdk.WindowTypeHint.DOCK)
        self.set_keep_below(True)

        screen = self.get_screen()
        visual = screen.get_rgba_visual()
        if visual != None and screen.is_composited():
            self.set_visual(visual)

        self.show_all()


    @property
    def width(self):

        return self.get_size()[0]


    @property
    def height(self):

        return self.get_size()[1]






class Gulik(object):

    """
    The main object thingamabob.
    """

    monitor_table = {
        'cpu': monitors.CPUMonitor,
        'memory': monitors.MemoryMonitor,
        'network': monitors.NetworkMonitor,
        'battery': monitors.BatteryMonitor,
        'disk': monitors.DiskMonitor
    }


    def __init__(self, configpath):

        self.started = False
        self.will_to_live = True # only needed because of gtk
        
        self.logo = cairo.ImageSurface.create_from_png(os.path.join(__path__[0], 'gulik.png'))

        self.screen = Gdk.Screen.get_default()
        self.window = Window()
        self.window.connect('draw', self.draw)

        self.configpath = configpath
        self.config = {}
        self.setup = self.autosetup

        self.monitors = {}
        self.visualizers = []
        #self.boxes = []

        self.apply_config()
        

    def reset(self):

        # clear out any existing visualizers and monitors
        for monitor in self.monitors.values(): # kill any existing monitors
            monitor.seppuku = True

        self.monitors = {}
        self.visualizers = []
        #self.boxes = []


    def module_to_config(self, locals):

        config = {}

        for key in DEFAULTS:
            config[key] = locals.get(key, DEFAULTS[key])

        for key in set(locals) - set(DEFAULTS): # iterates through everything defined in config.py we haven't already covered with DEFAULTS
            if key == key.upper(): # all-caps means it's config
                config[key] = locals[key]

        if not 'COLOR_FOREGROUND_TEXT' in config:
            config['COLOR_FOREGROUND_TEXT'] = config['COLOR_CAPTION']

        return config


    def add_netdata_monitors(self):

        # create monitors for all netdata hosts
        for host in self.config['NETDATA_HOSTS']:

            if isinstance(host, (list, tuple)): # handle (host, port) tuples and bare hostnames as values
                host, port = host
            else:
                port = 19999

            component = f"netdata-{host}"
            self.monitors[component] = monitors.NetdataMonitor(self, component, host, port)


    def resize_and_move(self):

        self.window.resize(self.config['WIDTH'], self.config['HEIGHT'])
        self.window.move(self.config['X'], self.config['Y']) # move apparently must be called after show_all


    def apply_config(self):

        print(f"Trying to load config from {self.configpath}…")

        custom = False # whether gulik is using a custom configuration
        old_config = self.config
        old_setup = self.setup

        self.config = {}
        try:
            fd = open(self.configpath, mode='r')
            config_string = fd.read()
        except OSError as e:
            print("No config at '%s' or insufficient permissions to read it. Falling back to defaults…" % self.configpath)
            config_locals = DEFAULTS

        else:
            try:
                config_locals = {}
                exec(config_string, config_locals)
                custom = True
            except Exception as e:
                print("Error in '%s': %s" % (self.configpath, e))
                print("Falling back to defaults…")
                config_locals = DEFAULTS

        self.config = self.module_to_config(config_locals)
        self.resize_and_move()

        if 'setup' in config_locals:
            self.setup = functools.partial(config_locals['setup'], app=self)
        else:
            self.setup = self.autosetup

        self.reset()

        self.add_netdata_monitors()


        # NOTE: psutil-based monitors are autoloaded in add_visualizer and thus don't have to be handled here like netdata monitors

        # finally, run the actual setup function to place all visualizers
        try:
            self.setup()
            print("Done.")

        except Exception as e:

            # Hokay, so I *think* this catches all failure scenarios for config (re)loads…
            
            if not custom:
                print("Sorry, friend. The autosetup seems to be broken on your machine. You might want to file a bug report.")
                print("Alternatively, you can create a custom-fit configuration at ~/.config/gulik/config.py.")
                self.stop()
                return

            else:

                if isinstance(self.setup, functools.partial):
                    name = self.setup.func.__name__
                else:
                    name = self.setup.__name__
                print(f"Your custom 'setup' function failed - {type(e).__name__}: {e}")

                print("Falling back to previous configuration")
                self.monitors = {}
                self.visualizers = []
                self.config = old_config or self.module_to_config(DEFAULTS) # DEFAULTS if config was empty before (happens on first time this function is called on a Gulik object
                self.setup = old_setup
                self.resize_and_move()
                self.add_netdata_monitors()

                try:
                    self.setup()
                except Exception as e:
                    print("Previous setup failed, too. Reverting to autosetup.")
                    self.monitors = {}
                    self.visualizers = []
                    self.config = self.module_to_config(DEFAULTS)
                    self.setup = self.autosetup
                    self.resize_and_move()
                    self.add_netdata_monitors()
                    try:
                        self.setup()
                    except Exception as e:
                        raise
                        print("Well, this should never happen.")
                        print(e)
                        self.stop()


    def signal_reload(self, *_):

        self.apply_config() # re-creates monitors and visualizers

        for monitor in self.monitors.values():
            monitor.start()


    def tick(self):

        for monitor in self.monitors.values():
            monitor.tick()
        self.window.queue_draw()
        return self.will_to_live # gtk stops executing timeout callbacks if they don't return True


    def draw_text(self, context, text, x, y, align=None, color=None, font_size=None, font_weight=None):
        
        if align is None:
            align = 'left_top'
        
        if color is None:
            color = self.config['COLOR_CAPTION']
        
        context.set_source_rgba(*color.tuple_rgba())

        font_size = font_size or self.config['FONT_SIZE']
        font_weight = font_weight or self.config['FONT_WEIGHT']

        font = Pango.FontDescription('%s %s %d' % (self.config['FONT'], font_weight, font_size))

        layout = PangoCairo.create_layout(context)
        layout.set_font_description(font)
        layout.set_text(text, -1)
        
        size = layout.get_pixel_size()

        x_offset, y_offset = helpers.alignment_offset(align, size)
        
        context.translate(x + x_offset, y + y_offset)

        PangoCairo.update_layout(context, layout)
        PangoCairo.show_layout(context, layout)

        context.translate(-x - x_offset, -y - y_offset)

        return size

    
    def draw_logo(self, context):

        width = self.logo.get_width()
        height = self.logo.get_height()
        
        context.save()
        context.set_operator(cairo.OPERATOR_SOFT_LIGHT)
        context.translate(self.config['WIDTH'] - width, self.config['HEIGHT'] - height)
        context.set_source_surface(self.logo)
        context.rectangle(0, 0, width, height)
        context.fill()
        context.restore()


    def draw(self, window, context):

        context.set_operator(cairo.OPERATOR_CLEAR)
        context.paint()
        context.set_operator(cairo.OPERATOR_OVER)

        context.set_source_rgba(*self.config['COLOR_WINDOW_BACKGROUND'].tuple_rgba())
        context.rectangle(0, 0, self.config['WIDTH'], self.config['HEIGHT'])
        context.fill()

        self.draw_logo(context)

        for visualizer in self.visualizers:
            visualizer.update(context)


    def start(self):

        assert self.config['FPS'] != 0
        assert isinstance(self.config['FPS'], float) or self.config['FPS'] >= 1

        for monitor in self.monitors.values():
            monitor.start()

        signal.signal(signal.SIGINT, self.stop) # so ctrl+c actually kills gulik
        signal.signal(signal.SIGTERM, self.stop) # so kill actually kills gulik, and doesn't leave a braindead Gtk.main and Collector processes dangling around
        signal.signal(signal.SIGUSR1, self.signal_reload) # reload config on user-defined signal. (no, not HUP)
        GLib.timeout_add(1000/self.config['FPS'], self.tick)
        self.tick() # first tick without delay so we get output asap
        self.started = True
        Gtk.main() # blocks until Gtk.main.quit is called
        print("\nThank you for flying with phryk evil mad sciences, LLC. Please come again.")


    def stop(self, num=None, frame=None):

        spinner = '▏▎▍▌▋▊▉▉▊▋▌▍▎'
        self.will_to_live = False # will stop calls by gtk main loop to self.tick

        for monitor in self.monitors.values():
            monitor.seppuku = True
            i = 0

            while monitor.is_alive(): # show spinner animation until the monitor is finished

                i += 1

                spinner_idx = i % len(spinner)

                sys.stdout.write(f"\r{spinner[spinner_idx]}    ")
                sys.stdout.flush()
                time.sleep(1/24)

            #sys.stdout.write("\rn")
            #sys.stdout.flush()

        if self.started: # Gtk.main.quit throws an error if called when not started
            Gtk.main_quit()
        else:
            exit() # to actually quit the program before Gtk.main is running


    def box(self, x=0, y=0, width=None, height=None): 

        width = helpers.ignore_none(width, self.config['WIDTH'] - x)
        height = helpers.ignore_none(height, self.config['HEIGHT'] - y)

        box = Box(self, x, y, width, height)
        #self.boxes.append(box)
        return box


    def add_visualizer(self, component, cls, **kwargs):

        if not component in self.monitors:
            if component in self.monitor_table:
                print("Autoloading %s!" % self.monitor_table[component].__name__)
                self.monitors[component] = self.monitor_table[component](self, component)
            elif component.startswith('netdata-'):
                raise LookupError(f"Unknown netdata host '{component[8:]}'")
            else:
                raise LookupError("No monitor class known for component '%s'. Custom monitor classes have to be added to Gulik.monitor_table to enable autoloading." % component)

        monitor = self.monitors[component]

            
        visualizer = cls(self, monitor, **kwargs)
        self.visualizers.append(visualizer)


    def autosetup(self, x=0, y=0, width=None, height=None):

        box = self.box(x, y, width, height)

        cpu_num = psutil.cpu_count()
        all_cores = ['core_%d' % x for x in range(0, cpu_num)]


        box.place(
            'cpu',
            visualizers.Arc,
            #elements=['aggregate'],
            elements=all_cores,
            width=box.width, 
            height=box.width,
            stroke_width=10,
            combination='cumulative_force',
            legend=False,
            padding_bottom=5,
            captions=[

                {
                    'text': 'CPU',
                    'position': 'left_top',
                    'align': 'left_top',
                    'font_weight': 'Bold',
                    'operator': Operator.HARD_LIGHT,
                },

                {
                    'text': '{aggregate:.1f}%',
                    'position': 'center_center',
                    'align': 'center_center',
                },
                {
                    'text': '{count} cores',
                    'position': 'right_bottom',
                    'align': 'right_bottom',
                }
            ]
        )

        box.place(
            'cpu',
            visualizers.Plot,
            elements=all_cores,
            width=box.width,
            height=130,
            autoscale=True,
            combination='cumulative_force',
        )

        box.place(
            'memory',
            visualizers.Arc,
            elements=['other', 'top_3', 'top_2', 'top_1'],
            width=box.width,
            height=box.width + 100,
            padding_bottom=80, # make space for 4 rows of legend
            stroke_width=30,
            combination='cumulative',
            #legend_format="{{{element}}}",
            legend_order='reverse',
            legend_format="{{{element}.name}} ({{{element}.size}})",
            captions=[
                {
                    'text': 'memory',
                    'position': 'left_top',
                    'align': 'left_top',
                    'font_weight': 'Bold',
                    'operator': Operator.HARD_LIGHT,
                },
                {
                    'text': '{percent:.1f}%',
                    'position': 'center_center',
                    'align': 'center_center'
                },

                {
                    'text': '{total}',
                    'position': 'right_bottom',
                    'align': 'right_bottom',
                }
            ]
        )

        last_visualizer = self.visualizers[-1]
        palette = [color for color in reversed(last_visualizer.palette(last_visualizer.colors['foreground'], 4))]

        all_nics = [x for x in psutil.net_if_addrs().keys()]
        all_nics_up = ['%s.bytes_sent' % x for x in all_nics]
        all_nics_down = ['%s.bytes_recv' % x for x in all_nics]
        all_nics_up_packets_ = ['%s.packets_sent' % x for x in all_nics]
        all_nics_down_packets = ['%s.packets_recv' % x for x in all_nics]
        all_nics_up_errors = ['%s.errout' % x for x in all_nics]
        all_nics_down_errors = ['%s.errin' % x for x in all_nics]
        all_nics_up_drop = ['%s.dropout' % x for x in all_nics]
        all_nics_down_drop = ['%s.dropin' % x for x in all_nics]

        partitions = psutil.disk_partitions()
        all_partitions = []
        for partition in partitions:
            name = partition.device.split('/')[-1].replace('.', '-')
            all_partitions.append(name)

        all_disks = list(psutil.disk_io_counters(perdisk=True))
        all_disks_read = [f"io.{disk}.read_bytes" for disk in all_disks]
        all_disks_write = [f"io.{disk}.write_bytes" for disk in all_disks]

        disk_busy_size = 130 / len(all_disks)

        box.place(
            'disk',
            visualizers.MirrorPlot,
            width=box.width - disk_busy_size,
            height=130,
            elements=[all_disks_read, all_disks_write],
            caption_placement='padding',
            captions=[
                {
                    'text': 'disk',
                    'position': 'left_top',
                    'align': 'left_top',
                    'font_weight': 'Bold',
                    'operator': Operator.HARD_LIGHT,
                }
            ]
        )

        last_visualizer = self.visualizers[-1]
        colors = last_visualizer.palette(last_visualizer.colors['foreground'], len(all_disks))
        for idx, disk in enumerate(all_disks):
            box.place(
                'disk',
                visualizers.Arc,
                width=disk_busy_size,
                height=disk_busy_size,
                foreground=colors[idx],
                legend=False,
                padding=0,
                stroke_width=10,
                elements=[f"io.{disk}.busy_time"]
            )

        box.place(
            'network',
            visualizers.MirrorPlot,
            width=box.width,
            height=140,
            margin_top=15,
            #padding=15,
            padding_bottom=50,
            elements=[all_nics_up, all_nics_down],
            combination='cumulative_force',
            autoscale=True,
            caption_placement='padding',
            captions=[
                {
                    'text': 'network',
                    'position': 'left_top',
                    'align': 'left_top',
                    'font_weight': 'Bold',
                    'operator': Operator.HARD_LIGHT,
                }
            ]
        )

        if psutil.sensors_battery() is not None:

            color = self.config['COLOR_FOREGROUND'].clone()
            color.alpha = 1

            box.place(
                'battery',
                visualizers.Rect,
                width=box.width,
                height=90,
                margin_top=35,
                elements=['battery'],
                foreground=color,
                captions=[
                    {
                        'text': '{percent}%',
                        'position': 'left_center',
                        'align': 'left_center',
                        'color': color,
                        'operator': Operator.DIFFERENCE, # such candy, many wow
                    },
                    {
                        'text': '{state}',
                        'position': 'right_center',
                        'align': 'right_center',
                        'color': color,
                        'operator': Operator.DIFFERENCE,
                    },
                ]
            )
