# -*- coding: utf-8 -*-

import os
import time
import signal
import psutil
import setproctitle
import multiprocessing

from . import helpers


class Collector(multiprocessing.Process):

    def __init__(self, app, queue_update, queue_data):

        super(Collector, self).__init__()
        self.daemon = True
        self.app = app
        self.queue_update = queue_update
        self.queue_data = queue_data
        self.elements = []


    def terminate(self):

        #self.queue_data.close() # closing queues manually actually seems to mess stuff up
       
        # Would've done this cleaner, but after half a day of chasing some
        # retarded quantenbug I'm done with this shit. Just nuke the fucking
        # things from orbit.
        os.kill(self.pid, signal.SIGKILL)
        #super(Collector, self).terminate()


    def run(self):

        setproctitle.setproctitle(f"gulik - {self.__class__.__name__}")

        while True:

            try:
                msg = self.queue_update.get(block=True)
                if msg == 'UPDATE':
                    self.update()

            except KeyboardInterrupt: # so we don't randomly explode on ctrl+c
                pass


    def update(self):
        raise NotImplementedError("%s.update not implemented!" % self.__class__.__name__)


class CPUCollector(Collector):

    def update(self):

        count = psutil.cpu_count()
        aggregate = psutil.cpu_percent(percpu=False)
        percpu = psutil.cpu_percent(percpu=True)
        self.queue_data.put(
            {
                'count': count,
                'aggregate': aggregate, 
                'percpu': percpu
            },
            block=True
        )
        
        # according to psutil docs, there should at least be 0.1 seconds
        # between calls to cpu_percent without sampling interval
        time.sleep(0.1) 


class MemoryCollector(Collector):

    def update(self):

        vmem = psutil.virtual_memory()

        processes = []
        total_use = 0
        for process in psutil.process_iter():

            if psutil.LINUX or (psutil.BSD and not self.app.config['BSD_ACCURATE_MEMORY']):

                if psutil.BSD:
                    key = 'rss'
                else:
                    key = 'pss'

                try:

                    pmem = process.memory_full_info()._asdict()
                    processes.append(helpers.DotDict({
                        'name': process.name(),
                        'size': pmem[key],
                        #'shared': pmem.shared,
                        'percent': pmem[key] / vmem.total * 100
                    }))

                    total_use += pmem[key]

                except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess) as e:
                    continue # skip to next process

            elif psutil.BSD:

                try:
                    resident = 0
                    size = 0
                    #shared = 0
                        
                    try:
                        for mmap in process.memory_maps():

                            # assuming everything with a real path is
                            # "not really in ram", but no clue.
                            if  mmap.path.startswith('['): 
                                size += mmap.private * PAGESIZE
                                resident += mmap.rss * PAGESIZE
                                #shared += (mmap.rss - mmap.private) * PAGESIZE # FIXME: probably broken, can yield negative values
                    except OSError:
                        pass # probably "device not available"

                    processes.append(helpers.DotDict({
                        'name': process.name(),
                        'size': size,
                        #'shared': resident - size,
                        'percent': size / vmem.total * 100,
                    }))
                    
                    total_use += size

                except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess) as e:
                    pass# TODO: add counter for processes we can't introspect

        info = helpers.DotDict({
            'total': vmem.total,
            'percent': total_use / vmem.total * 100,
            'available': vmem.total - total_use
        })

        processes_sorted = sorted(processes, key=lambda x: x['size'], reverse=True)

        for i, process in enumerate(processes_sorted[:3]):
            info['top_%d' % (i + 1)] = process 

        info['other'] = helpers.DotDict({
            'name': 'other',
            'size': 0,
            #'shared': 0,
            'count': 0
        })

        for process in processes_sorted[3:]:

            info['other']['size'] += process['size']
            #info['other']['shared'] += process['shared']
            info['other']['count'] += 1

        info['other']['percent'] = info['other']['size'] / vmem.total * 100

        self.queue_data.put(info, block=True)

        #time.sleep(1) # because this became horribly slow


class NetworkCollector(Collector):

    def update(self):

        stats = psutil.net_if_stats()
        addrs = psutil.net_if_addrs()
        counters = psutil.net_io_counters(pernic=True)
        connections = psutil.net_connections(kind='all')

        self.queue_data.put(
            {
                'stats': stats,
                'addrs': addrs,
                'counters': counters,
                'connections': connections,
            },
            block=True
        )


class BatteryCollector(Collector):

    def update(self):
        
        self.queue_data.put(psutil.sensors_battery(), block=True)


class DiskCollector(Collector):

    def __init__(self, *args, **kwargs):

        super(DiskCollector, self).__init__(*args, **kwargs)
        self.previous_io = {}

    def update(self):

        data = {}

        partitions = psutil.disk_partitions()

        data['partitions'] = {}
        for partition in partitions:

            name = partition.device.split('/')[-1].replace('.', '-')
            data['partitions'][name] = partition._asdict()
            data['partitions'][name]['name'] = name

            data['partitions'][name]['usage'] = psutil.disk_usage(partition.mountpoint)._asdict()
        
        io = psutil.disk_io_counters(perdisk=True)
        data['io'] = {}
        for disk, info in io.items():

            previous_info = self.previous_io.get(disk, None)

            if previous_info is None:

                data['io'][disk] = {
                    'read_count': 0,
                    'write_count': 0,
                    'read_bytes': 0,
                    'write_bytes': 0,
                    'read_time': 0,
                    'write_time': 0,
                    'busy_time': 0
                }

            else:

                info = info._asdict()
                previous_info = previous_info._asdict()

                data['io'][disk] = {}
                for k in ['read_count', 'write_count', 'read_bytes', 'write_bytes', 'read_time', 'write_time', 'busy_time']:
                    data['io'][disk][k] = (info[k] - previous_info[k]) * self.app.config['FPS']

            #data['io'][disk] = info._asdict()


        self.queue_data.put(data, block=True)
        self.previous_io = io


class NetdataCollector(Collector):

    def __init__(self, app, queue_update, queue_data, host, port):

        super(NetdataCollector, self).__init__(app, queue_update, queue_data)
        self.client = netdata.Netdata(host, port=port, timeout=1/self.app.config['FPS'])


    def run(self):

        setproctitle.setproctitle(f"gulik - {self.__class__.__name__}")

        while True:

            try:
                msg = self.queue_update.get(block=True)
                if msg.startswith('UPDATE '):
                    chart = msg[7:]
                    self.update(chart)

            except KeyboardInterrupt: # so we don't randomly explode on ctrl+c
                pass


    def update(self, chart):

        try:
            # get the last second of data condensed to one point
            data = self.client.data(chart, points=1, after=-1, options=['absolute'])
        except netdata.NetdataException:
            pass
        else:
            self.queue_data.put((chart, data), block=True)


